﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Windows.Forms;
using Microsoft.WindowsAPICodePack.Dialogs;
using System.IO;
using System.Diagnostics;

namespace ResizeAll
{
    public partial class Form1 : Form
    {
        /// <summary>
        /// Standard date format. This just specifies how dates are represented in the status log.
        /// https://msdn.microsoft.com/en-us/library/zdtaw1bw(v=vs.110).aspx
        /// y = year, M = month, d = day, H = hour, m = minute, s = seconds, f = milliseconds
        /// </summary>
        public const string cStandardDateFormat = "yyyy-MM-dd HH:mm:ss.ffff";

        /// <summary>
        /// Files with these extensions are considered image files by the program.
        /// </summary>
        private static readonly List<string> cImageFileExtensions = new List<string>()
        {
            ".PNG",
            ".JPG",
            ".JPEG",
            ".BMP",
            ".TIF",
            ".TIFF"
        };

        /// <summary>
        /// A delegate linking the static PostMessage method to the non-static WriteMessageLine method.
        /// </summary>
        private static Action<string> writeMessageHook;

        public Form1()
        {
            this.InitializeComponent();

            // Since there's only one main form, this is a little trick I use to allow the main form to handle 
            // messages from other methods as if the main form were a static class.

            if(writeMessageHook != null)
            {
                // There can only be one hook responding to event messages.
                MessageBox.Show("Attempted to create more than one main form. The program will now exit.");
                Application.Exit();
            }

            // hook this form's static event handler into the instance handler (hacky, but convenient)
            writeMessageHook = WriteMessageLine;
        }
        

        private void btnBrowseInputDirectory_Click(object sender, EventArgs e)
        {
            string selectedDirectory;

            using(CommonOpenFileDialog dlgSelectFolder = new CommonOpenFileDialog())
            {
                dlgSelectFolder.IsFolderPicker = true;

                CommonFileDialogResult result = dlgSelectFolder.ShowDialog();

                if(result != CommonFileDialogResult.Ok)
                {
                    return;
                }

                selectedDirectory = dlgSelectFolder.FileName;
            }

            txtImagesFolder.Text = selectedDirectory;

            // make it so that the default output directory is a subfolder in the input directory called "Resized"
            txtOutputDirectory.Text = Path.Combine(selectedDirectory, "Resized");
        }

        private void btnBrowseOutputDirectory_Click(object sender, EventArgs e)
        {
            string selectedDirectory;

            using(CommonOpenFileDialog dlgSelectFolder = new CommonOpenFileDialog())
            {
                dlgSelectFolder.IsFolderPicker = true;

                CommonFileDialogResult result = dlgSelectFolder.ShowDialog();

                if(result != CommonFileDialogResult.Ok)
                {
                    return;
                }

                selectedDirectory = dlgSelectFolder.FileName;
            }

            txtOutputDirectory.Text = selectedDirectory;
        }


        private void btnStart_Click(object sender, EventArgs e)
        {
            string selectedDirectory = txtImagesFolder.Text;

            if(string.IsNullOrWhiteSpace(selectedDirectory))
            {
                MessageBox.Show("Please select a directory of images to resize.");
                return;
            }

            string[] filesInDirectory;

            // It is possible for Windows to deny us access to list the files in a folder. Unlikely but very possible.
            try
            {
                filesInDirectory = Directory.GetFiles(selectedDirectory);
            }
            catch(Exception ex)
            {
                MessageBox.Show($"Failed to list files in directory '{selectedDirectory}': {ex.Message}");
                return;
            }

            // Go through the files in the directory and find the ones that are image files
            List<string> imageFilePaths = new List<string>();

            foreach(string filePath in filesInDirectory)
            {
                if(IsImageFile(filePath))
                {
                    imageFilePaths.Add(filePath);
                }
            }

            PostMessage($"Detected {imageFilePaths.Count} images in directory '{selectedDirectory}'.");

            string outputDirectory = txtOutputDirectory.Text;
            float scaleFactor = (float) nudScaleFactor.Value;

            // Process each image file
            foreach(string imageFilePath in imageFilePaths)
            {
                ScaleImage(imageFilePath, outputDirectory, scaleFactor);
            }

            PostMessage("Image resizing operation complete.");
        }

        /// <summary>
        /// Reads in an input image, scales it by scaleFactor, and writes the scaled image to the output directory, with "Resized_" at the beginning of the file name.
        /// 
        /// Does not throw exceptions, it logs all errors.
        /// </summary>
        private static void ScaleImage(string inputFilePath, string outputDirectory, float scaleFactor)
        {
            Image imageFromFile = null;

            try
            {
                string inputFileName;

                try
                {
                    imageFromFile = Bitmap.FromFile(inputFilePath);

                    inputFileName = Path.GetFileName(inputFilePath);

                    if(imageFromFile == null)
                    {
                        throw new IOException("No image data in file.");
                    }

                    if(!(imageFromFile is Bitmap))
                    {
                        throw new IOException("File did not contain a valid Bitmap image.");
                    }
                }
                catch(Exception ex)
                {
                    // You must always be prepared for file loading to fail.
                    //
                    // here's some examples of how Bitmap.FromFile could fail:
                    // - The file got deleted (you could check to see if it exists but it could get deleted the nanosecond after you checked, so there's no point)
                    // - The file isn't there for some other reason (flash drive got ejected, network drive connection lost)
                    // - You don't have read access to that file (you can check if you have rights, but admins and the OS can change this at any nanosecond)
                    // - The file isn't an image file (rename some random text document to TestFile.BMP)
                    // - The file is an image file, but isn't one that Bitmap supports (try it with a wmf or svg file)
                    throw new IOException($"Failed to load input image from '{inputFilePath}': {ex.Message}", ex);
                }

                int oldWidth    = imageFromFile.Width;
                int oldHeight   = imageFromFile.Height;

                // The Max is in there just so that we don't somehow get an image width or height of 0 pixels which wouldn't be good...
                int newWidth    = Math.Max(1, (int) (oldWidth   * scaleFactor));
                int newHeight   = Math.Max(1, (int) (oldHeight  * scaleFactor));

                string outputFileName = $"Resized_{inputFileName}";

                string outputFilePath = Path.Combine(outputDirectory, outputFileName);

                // it's possible that the directory we're saving to does not exist, so we need to create it, oterwise outputImage.Save will fail.
                
                try
                {
                    Directory.CreateDirectory(outputDirectory);
                }
                catch(Exception ex)
                {
                    // You must always be prepared for directory creation to fail.
                    // 
                    // - It's possible that the OS will deny us the ability to create a directory (you can check if you have rights, but admins and the OS can change this at any nanosecond)
                    throw new IOException($"Failed to create output directory '{outputDirectory}': {ex.Message}", ex);
                }

                try
                {
                    using(Bitmap outputImage = new Bitmap(imageFromFile, new Size(newWidth, newHeight)))
                    {
                        outputImage.Save(outputFilePath);
                        PostMessage($"Done processing image '{outputFilePath}'.");
                    }
                }
                catch(Exception ex)
                {
                    // You must always be prepared for file saving to fail.
                    //
                    // Here's some examples of how outputImage.Save could fail:
                    // - The output directory got deleted since we created it
                    // - We don't have write access to that directory
                    // - The file already exists and we don't have access to replace it
                    // - While writing the file, we lost access to it (flash drive got yanked, network drive connection lost, floppy disk got ejected)
                    // - The computer ran out of hard disk space (you can check, but the hard drive could be full again immediately afterwards)
                    throw new IOException($"Failed to save scaled image to '{outputFilePath}': {ex.Message}", ex);
                }

            }
            catch(Exception ex)
            {
                PostError($"Unable to resize file '{inputFilePath}': {ex.Message}");
                return;
            }
            finally
            {
                if(imageFromFile != null)
                {
                    imageFromFile.Dispose();
                }
            }
        }

        private static bool IsImageFile(string filePath)
        {
            string extension = Path.GetExtension(filePath).ToUpperInvariant();

            if(cImageFileExtensions.Contains(extension))
            {
                return true;
            }
            else
            {
                return false;
            }
        }


        /// <summary>
        /// Writes a line of text to the message log (rich text box), with the current date and time.
        /// </summary>
        private void WriteMessageLine(string message)
        {
            string textToWrite = $"{DateTime.Now.ToString(cStandardDateFormat)}: {message}{Environment.NewLine}";
            rtbMessages.AppendText(textToWrite);
        }

        private static void PostMessage(string message)
        {
            if(writeMessageHook != null)
            {
                writeMessageHook(message);
            }
        }

        private static void PostError(string message, Exception ex = null)
        {
            StackTrace callStack = new StackTrace();

            PostMessage($"[Error] : {message}");
            PostMessage($"Call Stack: {callStack.ToString()}");

            if(ex != null)
            {
                PostMessage($"Exception info: {ex.ToString()}");
            }
        }



        private void btnClearLog_Click(object sender, EventArgs e)
        {
            rtbMessages.Clear();
        }
    }
}
